from . import abc, sim_info, trees, tree_branches, groups, snapshots
from .sim_info import SimInfo, _Predefined, predefined
from .trees import (TreeLoader, TreeLoaderEagle, TreeLoaderSimba, 
                    TreeLoaderElucidExt, TreeLoaderElucidExtV2,
                    TreeLoaderTng, TreeLoaderTngDark)
from .groups import (SubhaloLoader, SubhaloLoaderTng, SubhaloLoaderEagle)
from .snapshots import (SnapshotLoader, SnapshotLoaderDmo, SnapshotLoaderTngDark, 
                        SnapshotLoaderTngDark, SnapshotLoaderBaryonic, SnapshotLoaderTng)


from __future__ import annotations
from pyhipp.io import h5
from pyhipp.core import abc
from .sim_info import SimInfo
from typing import Dict
import numpy as np

class SubhaloLoader(abc.HasLog, abc.HasCache):
    '''
    grp_id              -- offset into the group catalog at the same snapshot.
    '''
    def __init__(self, sim_info: SimInfo, snap: int, **base_kw) -> None:
        super().__init__(**base_kw)
        
        self.sim_info = sim_info
        self.snap = snap
    
        self.n_subhs: int = None
        self.n_grps: int = None
        self.root_file: h5.File = None
        self.dgrp_subh: h5.Group = None
        self.dataset_key_map: Dict[str, (str, int, float)] = {}
        
        self._preproc()
    
    def __getitem__(self, key):
        if not isinstance(key, str):
            return tuple(self[k] for k in key)
        return self.get_cache_or(key, lambda : self._load(key) )
    
    @staticmethod
    def create(sim_info: SimInfo, **kw):
        cls = {
            'tng': SubhaloLoaderTng,
            'tng_dark': SubhaloLoaderTng,
            'eagle': SubhaloLoaderEagle,
        }[sim_info.name]
        return cls(sim_info, **kw)
    
    def _preproc(self):
        raise NotImplementedError()
    
    def _load(self, key) -> np.ndarray:
        val = self.dataset_key_map[key]
        if isinstance(val, tuple):
            key_used, col_slice, scale = val
            out = self._load_impl(self.dgrp_subh, key_used, 
                               col_slice=col_slice, scale=scale)
        else:
            out = val()
        return out

    def _load_impl(self, g: h5.Group, key, col_slice = None, scale = None):
        '''
        @key: str | tuple of keys. 
        For the later, col_slice and scale are applied to the value of each 
        key; returned values are stacked.
        '''
        if isinstance(key, str):
            d = g[key]
            val = d[()] if col_slice is None else d[:, col_slice]
            if scale is not None:
                val *= scale
            return val

        val = np.stack(
            tuple( self._load_impl(g, k, col_slice, scale) for k in key ), 
            axis=-1)
        return val   
    
class SubhaloLoaderTng(SubhaloLoader):
    '''
    For TNGDark, no m_star and sfr.
    '''
    def _preproc(self):
        info = self.sim_info
        
        file = h5.File(info.root_dir / info.root_file)
        g_base = file[f'Groups/{self.snap}']
        head = g_base['Header'].attrs
        g_subh = g_base['Subhalo']
        g_grp = g_base['Group']
        
        z = self.sim_info.redshifts[self.snap]
        _a = 1.0 / (1.0 + z)
        r_scale = 1.0e-3 * _a
        
        n_subhs, n_grps = head['Nsubgroups_Total', 'Ngroups_Total']
        key_map = {
            'grp_id': ('SubhaloGrNr', None, None),
            'm_gas': ('SubhaloMassInRadType', 0, None),
            'm_star': ('SubhaloMassInRadType', 4, None),
            'm_bh_': ('SubhaloMassInRadType', 5, None),   # including attached gas
            'm_bh': ('SubhaloBHMass', None, None),        # real BH mass
            
            'n_p_gas': ('SubhaloLenType', 0, None),       # number of particles/cells  
            'n_p_dm': ('SubhaloLenType', 1, None),
            'n_p_star': ('SubhaloLenType', 4, None),
            'n_p_bh': ('SubhaloLenType', 5, None),
            
            'sfr': ('SubhaloSFRinRad', None, None),
            'x': ('SubhaloPos', None, 1.0e-3),                         # cMpc/h
            'v': ('SubhaloVel', None, None),
            
            'r_half_gas': ('SubhaloHalfmassRadType', 0, r_scale),      # pMpc/h
            'r_half_dm': ('SubhaloHalfmassRadType', 1, r_scale),
            'r_half_star': ('SubhaloHalfmassRadType', 4, r_scale),
            
            'flag': self._load_flag,  
            'is_c': self._load_is_cent,
            
            'snap_off_gas': lambda : self._load_offset('snap_off_gas'),
            'snap_off_dm': lambda : self._load_offset('snap_off_dm'),
            'snap_off_star': lambda: self._load_offset('snap_off_star'),
            'id_in_tree': lambda : self._load_offset('id_in_tree'),
            'subhalo_id': lambda : self._load_offset('subhalo_id'),
            'last_pro': lambda : self._load_offset('last_pro'),
            
            'grp_m_tophat': lambda : self._load_grp('m_tophat'),
            'grp_m_mean200': lambda : self._load_grp('m_mean200'),
            'grp_m_crit200': lambda : self._load_grp('m_crit200'),
            'grp_x': lambda : self._load_grp('x'),
        }
        
        self.n_subhs = n_subhs
        self.n_grps = n_grps
        self.root_file = file
        self.dgrp_subh = g_subh
        self.dgrp_grp = g_grp
        self.dataset_key_map = key_map
        
    def _load_flag(self):
        if self.sim_info.is_baryon:
            out = self.dgrp_subh.datasets['SubhaloFlag']
            out = np.asarray(out, dtype=bool)
        else:
            out = np.ones(self.n_subhs, dtype=bool)
        return out
    
    def _load_is_cent(self):
        cid = self.root_file[f'Groups/{self.snap}/Group'].datasets['GroupFirstSub']
        is_c = np.zeros(self.n_subhs, dtype=bool)
        cid = cid[cid >= 0]
        is_c[cid] = True
        return is_c
    
    def _load_offset(self, key):
        offcat = self.root_file[f'Offsets/{self.snap}/Subhalo']
        subl_dsets = offcat['SubLink'].datasets
        if key == 'snap_off_gas':
            out = offcat['SnapByType'][:, 0]
        elif key == 'snap_off_dm':
            out = offcat['SnapByType'][:, 1]
        elif key == 'snap_off_star':
            out = offcat['SnapByType'][:, 4]
        elif key == 'id_in_tree':
            out = subl_dsets['RowNum']
        elif key == 'subhalo_id':
            out = subl_dsets['SubhaloID']
        elif key == 'last_pro':
            out = subl_dsets['LastProgenitorID']
        else:
            raise KeyError(key)
        return out
    
    def _load_grp(self, key):
        grp_id = self['grp_id']
        
        g = self.dgrp_grp
        if key == 'x':
            out = g['GroupPos'][()] * 1.0e-3
        elif key == 'm_tophat':
            out = g['Group_M_TopHat200'][()]
        elif key == 'm_mean200':
            out = g['Group_M_Mean200'][()]
        elif key == 'm_crit200':
            out = g['Group_M_Crit200'][()]
        else:
            raise KeyError(key)
        
        return out[grp_id]
    
class SubhaloLoaderEagleRaw(abc.HasDictRepr):
    
    def __init__(self, sim_info: SimInfo, snap: int, **base_kw) -> None:
        super().__init__(**base_kw)
        
        path = sim_info.root_dir / sim_info.root_file
        file = h5.File(path)
        
        h = sim_info.cosmology.hubble
        m_scale = 1.0e-10 * h
        x_scale_mpc = h
        x_scale_kpc = 1.0e-3 * h
        apert_list = [1,3,5,10,20,30,40,50,70,100]      # pkpc
        prop_key_map = {
            'z': ('sub', 'Redshift', (), None),
            'snap': ('sub', 'SnapNum', (), None),
            'spurious': ('sub', 'Spurious', (), None),
            
            'des': ('sub', 'DescendantID', (), None),
            'last_pro': ('sub', 'LastProgID', (), None),
            'main_leaf': ('sub', 'TopLeafID', (), None),
            'subhalo_id': ('sub', 'GalaxyID', (), None),
            'group_id': ('sub', 'GroupID', (), None),
            'group_no': ('sub', 'GroupNumber', (), None),
            'subgroup_no': ('sub', 'SubGroupNumber', (), None),
            
            'x_com': ('sub', 'CentreOfMass', ('_x', '_y', '_z'), x_scale_mpc),  # cMpc/h
            'x': ('sub', 'CentreOfPotential', ('_x', '_y', '_z'), x_scale_mpc), # cMpc//h
            'v': ('sub', 'Velocity', ('_x', '_y', '_z'), None),                 # peculiar velocity, km/s
            
            
            'mass': ('sub', 'Mass', (), m_scale),                               # 1.0e10 Msun/h
            'mass_type_bh': ('sub', 'MassType_BH', (), m_scale),
            'mass_type_dm': ('sub', 'MassType_DM', (), m_scale),
            'mass_type_gas': ('sub', 'MassType_Gas', (), m_scale),
            'mass_type_star': ('sub', 'MassType_Star', (), m_scale),
            'mass_nsf': ('sub', 'NSF_Mass', (), m_scale),
            'mass_sf': ('sub', 'SF_Mass', (), m_scale),
            
            'sfr': ('sub', 'StarFormationRate', (), None),                      # Msun/yr
            'vel_disp_star': ('sub', 'StellarVelDisp', (), None),               # km/s
            'v_max': ('sub', 'Vmax', (), None),                                 # km/s
            'r_v_max': ('sub', 'VmaxRadius', (), x_scale_kpc),                  # pMpc/h
            
            'm_bh': ('sub', 'BlackHoleMass', (), m_scale),
            'm_bh_dot': ('sub', 'BlackHoleMassAccretionRate', (), None),        # Msun/yr
            
            'spin_gas': ('sub', 'GasSpin', ('_x', '_y', '_z'), x_scale_kpc),    # pMpc/h km/s
            'spin_nsf': ('sub', 'NSF_Spin', ('_x', '_y', '_z'), x_scale_kpc),
            'spin_sf': ('sub', 'SF_Spin', ('_x', '_y', '_z'), x_scale_kpc),
            'spin_star': ('sub', 'Stars_Spin', ('_x', '_y', '_z'), x_scale_kpc),
            
            'r_half_bh': ('sub', 'HalfMassRad_BH', (), x_scale_kpc),            # pMpc/h
            'r_half_dm': ('sub', 'HalfMassRad_DM', (), x_scale_kpc),
            'r_half_gas': ('sub', 'HalfMassRad_Gas', (), x_scale_kpc),
            'r_half_star': ('sub', 'HalfMassRad_Star', (), x_scale_kpc),
            'r_half_proj_bh': ('sub', 'HalfMassProjRad_BH', (), x_scale_kpc),
            'r_half_proj_dm': ('sub', 'HalfMassProjRad_DM', (), x_scale_kpc),
            'r_half_proj_gas': ('sub', 'HalfMassProjRad_Gas', (), x_scale_kpc),
            'r_half_proj_star': ('sub', 'HalfMassProjRad_Star', (), x_scale_kpc),
            
            'z_fof': ('grp', 'Redshift', (), None),
            'snap_fof': ('grp', 'SnapNum', (), None),
            
            'id_fof': ('grp', 'GroupID', (), None),
            'first_sub_fof': ('grp', 'GroupFirstSub', (), None),                # 0-indexed, offset
            'n_sub_fof': ('grp', 'NumOfSubhalos', (), None),
            
            'x_fof': ('grp', 'GroupCentreOfPotential', ('_x', '_y', '_z'), x_scale_mpc),
            'm_fof': ('grp', 'GroupMass', (), m_scale),
            
            'm_crit_200': ('grp', 'Group_M_Crit200', (), m_scale),
            'm_crit_500': ('grp', 'Group_M_Crit500', (), m_scale),
            'm_crit_2500': ('grp', 'Group_M_Crit2500', (), m_scale),
            'm_mean_200': ('grp', 'Group_M_Mean200', (), m_scale),
            'm_mean_500': ('grp', 'Group_M_Mean500', (), m_scale),
            'm_mean_2500': ('grp', 'Group_M_Mean2500', (), m_scale),
            'm_tophat': ('grp', 'Group_M_TopHat200', (), m_scale),
            'r_crit_200': ('grp', 'Group_R_Crit200', (), x_scale_kpc),          # pMpc/h
            'r_crit_500': ('grp', 'Group_R_Crit500', (), x_scale_kpc),
            'r_crit_2500': ('grp', 'Group_R_Crit2500', (), x_scale_kpc),
            'r_mean_200': ('grp', 'Group_R_Mean200', (), x_scale_kpc),
            'r_mean_500': ('grp', 'Group_R_Mean500', (), x_scale_kpc),
            'r_mean_2500': ('grp', 'Group_R_Mean2500', (), x_scale_kpc),
            'r_tophat': ('grp', 'Group_R_TopHat200', (), x_scale_kpc),
        }
        for apert in apert_list:
            ap_key = f'ap{apert}'
            prop_key_map |= {
                f'sfr_{ap_key}': (ap_key, 'SFR', (), None),
                f'mass_{ap_key}_star': (ap_key, 'Mass_Star', (), m_scale),
                f'mass_{ap_key}_gas': (ap_key, 'Mass_Gas', (), m_scale),
                f'mass_{ap_key}_dm': (ap_key, 'Mass_DM', (), m_scale),
                f'mass_{ap_key}_bh': (ap_key, 'Mass_BH', (), m_scale),
                f'vel_disp_star_{ap_key}': (ap_key, 'VelDisp', (), None),
            }
        
        self.sim_info = sim_info
        self.snap = snap
        self.file = file
        self.apert_list = apert_list
        self.prop_key_map = prop_key_map
        
    def __getitem__(self, *keys):
        if len(keys) == 0:
            return None
        if len(keys) == 1:
            return self.__load_one(keys[0])
        return tuple(self.__load_one(k) for k in keys)
        
    def __load_one(self, key: str):
        cat, dset, comps, scale = self.prop_key_map[key]
        
        if cat == 'sub':
            g = self.subhalos
        elif cat == 'grp':
            g = self.groups
        elif cat[:2] == 'ap':
            g = self.subhalos_apertured(cat[2:])
        else:
            raise KeyError(cat)
        
        n_comps = len(comps)
        if n_comps == 0:
            val = g.datasets[dset]
        else:
            val = np.column_stack([g.datasets[dset + comp] for comp in comps])
        
        if scale is not None:
            val = val * scale
            
        return val
        
    def __enter__(self):
        return self
    
    def __exit__(self, *args, **kw_args):
        self.file.close()
        
    @property
    def snap_data_group(self):
        return self.file[f'Groups/{self.snap}']
        
    @property
    def subhalos(self):
        return self.snap_data_group['SubHalo']
        
    @property
    def groups(self):
        return self.snap_data_group['FOF']
    
    def subhalos_apertured(self, size = '30'):
        return self.snap_data_group[f'Aperture/{size}']
    
class SubhaloLoaderEagle(SubhaloLoader):
    def __init__(self, sim_info: SimInfo, snap: int, aperture='30', **base_kw) -> None:
        self.aperture = aperture
        super().__init__(sim_info, snap, **base_kw)
    
    def _preproc(self):
        info, snap, aperture = self.sim_info, self.snap, self.aperture
        
        file = h5.File(info.root_dir / info.root_file)
        g_snap = file[f'Groups/{snap}']
        g_subh, g_aperture, g_grp = g_snap['SubHalo', f'Aperture/{aperture}', 'FOF']
        
        n_grps, n_subhs = g_snap['Header'].attrs['NumFOFGroups', 'NumSubhalos']
        m_cvt_scale = 1.0e-10 * info.cosmology.hubble
        
        key_map = {
            'grp_id': lambda: g_subh['GroupNumber'][()] - 1,
            'm_star': lambda: g_aperture['Mass_Star'][()] * m_cvt_scale,
            'm_bh_': lambda: g_aperture['Mass_BH'][()] * m_cvt_scale,
            'm_bh': lambda: g_subh['BlackHoleMass'][()] * m_cvt_scale,
            'sfr': lambda: g_aperture['SFR'][()],
            
            'flag': lambda: (g_subh['Spurious'][()] == 0),
            'is_c': self._load_is_cent,
            
            'grp_cid': lambda: g_grp['GroupFirstSub'][()],
            'grp_n_subs': lambda: g_grp['NumOfSubhalos'][()][self['grp_id']],
            'grp_m_tophat': lambda: g_grp['Group_M_TopHat200'][()][self['grp_id']] * m_cvt_scale,
            'grp_m_mean200': lambda: g_grp['Group_M_Mean200'][()][self['grp_id']] * m_cvt_scale,
            'grp_m_crit200': lambda: g_grp['Group_M_Crit200'][()][self['grp_id']] * m_cvt_scale,
        }
        
        self.n_subhs = n_subhs
        self.n_grps = n_grps
        self.root_file = file
        self.dgrp_subh = g_subh
        self.dgrp_aperture = g_aperture
        self.dgrp_grp = g_grp
        self.dataset_key_map = key_map
        
    def _load_is_cent(self):
        grp_cid = self['grp_cid']
        grp_cid = grp_cid[grp_cid >= 0]
        
        is_c = np.zeros(self.n_subhs, dtype=bool)
        is_c[grp_cid] = True
        
        return is_c


from __future__ import annotations
import typing
from typing import Self, Literal, TypedDict, Unpack, Mapping
from pyhipp.core import abc, DataDict, DataTable
from .sim_info import SimInfo, predefined as predefined_sim_info
from pyhipp.stats import Rng
import numpy as np
from numpy.typing import NDArray
from pyhipp.io import h5
from pathlib import Path

_Value = int | float | bool | np.ndarray | None
_Range = tuple[_Value, _Value]
_ValueOpt = TypedDict('_ValueOpt', {
    'by': Literal['value', 'v'],
    'lo': _Value,
    'hi': _Value,
    'eq': _Value}, total=False)
_PercentileOpt = TypedDict('_PercentileOpt', {
    'by': Literal['percentile', 'p'],
    'p_lo': _Value, 'p_hi': _Value}, total=False)
_SelectOpt = dict[str, _Value | _Range | _ValueOpt | _PercentileOpt]


class HomogeneousObjectCatalog(abc.HasDictRepr, h5.abc.GroupLike):

    repr_attr_keys = ('n_objs', 'keys', 'header')
    h5_data_typedict = {
        'objs': DataTable,
        'header': DataDict
    }
    shallow_copy_attrs = ('header', )

    def __init__(self, objs: DataTable, header: DataDict, **base_kw) -> None:
        super().__init__(**base_kw)

        self.objs = objs
        self.header = header

    @property
    def keys(self) -> tuple[str, ...]:
        return tuple(self.objs.keys())

    @property
    def n_objs(self) -> int:
        for v in self.objs.values():
            return len(v)

    def __getitem__(self,
                    keys: str | tuple[str]
                    ) -> np.ndarray | tuple[np.ndarray]:
        return self.objs[keys]

    def select_by(self, opt_dict: _SelectOpt = {}) -> NDArray[np.bool_]:
        sel = np.ones(self.n_objs, dtype=bool)
        for key, opt in opt_dict.items():
            if isinstance(opt, tuple):
                lo, hi = opt
                sel &= self.select_by_value(key, lo=lo, hi=hi)
            elif isinstance(opt, Mapping):
                opt = {**opt}
                by = opt.pop('by')
                if by in ('value', 'v'):
                    sel &= self.select_by_value(key, **opt)
                elif by in ('percentile', 'p'):
                    sel &= self.select_by_percentile(key, **opt)
                else:
                    raise ValueError(f'Invalid option "by": {by}')
            else:
                sel &= self.select_by_value(key, eq=opt)

        return sel

    def select_by_value(self, key: str, lo=None, hi=None, eq=None
                        ) -> NDArray[np.bool_]:
        val, n_objs = self.objs[key], self.n_objs
        sel = np.ones(n_objs, dtype=bool)
        if lo is not None:
            sel &= val >= lo
        if hi is not None:
            sel &= val < hi
        if eq is not None:
            sel &= val == eq
        return sel

    def select_by_percentile(
            self, key: str, p_lo=None, p_hi=None) -> NDArray[np.bool_]:
        val = self.objs[key]
        if p_lo is None or p_lo <= 0.:
            lo = None
        else:
            lo = np.quantile(val, p_lo)

        if p_hi is None or p_hi >= 1.:
            hi = None
        else:
            hi = np.quantile(val, p_hi)

        return self.select_by_value(key, lo=lo, hi=hi)

    def shallow_copy(self, **init_kw) -> Self:
        init_kw = {
            k: getattr(self, k) for k in self.shallow_copy_attrs
        } | init_kw
        return type(self)(**init_kw)

    def subset(self, args: np.ndarray, **init_kw) -> Self:
        objs = self.objs.subset(args)
        return self.shallow_copy(objs=objs, **init_kw)

    def subset_by(self, opt_dict: _SelectOpt = {}, **init_kw) -> Self:
        '''
        Select a subset of objects by the combination of multiple criteria, 
        each applied to a column and then AND-ed.
        
        The key of `opt_dict` is the column name, and the value is one of 
        the following:
        - a tuple of two values, `(lo, hi)`.
        - a single value, `eq`.
        - a dictionary, with `by` indicating the method and the rest passed 
          as keyword arguments to the method.
          
        Examples
        --------
        
        ```py
        cat = HomogeneousObjectCatalog(...)
        cat.subset_by({
            'is_central': True,
            'm_h': (1.0e12, 1.0e13),
            'm_bh': (1.0e6, None),
            'm_s': {'by': 'p', 'p_lo': 0.1, 'p_hi': 0.9},
            'spin': {'by': 'v', 'lo': 0.1, 'hi': 0.5},
        })
        ```
        '''
        sel = self.select_by(opt_dict)
        return self.subset(sel, **init_kw)

    def subset_by_value(
            self, key: str, lo=None, hi=None, eq=None, **init_kw) -> Self:
        sel = self.select_by_value(key, lo=lo, hi=hi, eq=eq)
        return self.subset(sel, **init_kw)

    def subset_by_percentile(self, key: str, p_lo=None, p_hi=None,
                             **init_kw) -> Self:
        sel = self.select_by_percentile(key, p_lo=p_lo, p_hi=p_hi)
        return self.subset(sel, **init_kw)


class SimContext(abc.HasDictRepr, h5.abc.GroupLike):

    repr_attr_keys = ('sim_name', )

    def __init__(self, sim_info: SimInfo, rng: Rng.Initializer = 0) -> None:

        self.rng = Rng(rng)
        self.sim_info = sim_info

    @property
    def sim_name(self) -> str:
        return self.sim_info.name

    def _to_h5_data(self) -> dict:
        return super()._to_h5_data() | {
            'sim_name': self.sim_name
        }

    @classmethod
    def _from_h5_data(cls, data: dict, **init_kw) -> Self:
        sim_info = predefined_sim_info[data['sim_name'].decode()]
        return cls(sim_info, **init_kw)


class SimObjectCatalog(HomogeneousObjectCatalog):

    repr_attr_keys = HomogeneousObjectCatalog.repr_attr_keys + (
        'ctx', )
    h5_data_typedict = HomogeneousObjectCatalog.h5_data_typedict | {
        'ctx': SimContext}
    shallow_copy_attrs = HomogeneousObjectCatalog.shallow_copy_attrs + (
        'ctx',)

    def __init__(self, objs: DataTable, header: DataDict, ctx: SimContext,
                 **base_kw) -> None:
        super().__init__(objs=objs, header=header, **base_kw)

        self.ctx = ctx

    @property
    def rng(self) -> Rng:
        return self.ctx.rng

    @property
    def sim_info(self) -> SimInfo:
        return self.ctx.sim_info



from __future__ import annotations
import typing
from typing import Self
from pyhipp.core import abc
from .trees import TreeLoader
import numpy as np

class Branch(abc.HasDictRepr):

    repr_attr_keys = ('root_off', 'n_hs')

    def __init__(self, tr_ld: TreeLoader,
                 root_off: float,
                 n_hs: int = -1,
                 **kw) -> None:
        super().__init__(**kw)

        if n_hs == -1:
            hids, leaf_hids = tr_ld['subhalo_id', 'main_leaf']
            n_hs = leaf_hids[root_off] - hids[root_off] + 1

        self.root_off = root_off
        self.n_hs = n_hs
        self.tr_ld = tr_ld

    def val(self, *keys):
        assert len(keys) > 0
        b, n = self.root_off, self.n_hs
        e = b + n
        if len(keys) == 1:
            key = keys[0]
            out = self.tr_ld[key][b:e][::-1].copy()
        else:
            out = tuple(self.val(key) for key in keys)
        return out

    def root_val(self, *keys):
        assert len(keys) > 0
        b = self.root_off
        if len(keys) == 1:
            key = keys[0]
            out = self.tr_ld[key][b]
        else:
            out = tuple(self.root_val(key) for key in keys)
        return out


class BranchSet(abc.HasDictRepr):

    repr_attr_keys = ('size',)

    def __init__(self, tr_ld: TreeLoader,
                 branches: list[Branch],
                 **kw) -> None:
        super().__init__(**kw)

        self.branches = branches
        self.tr_ld = tr_ld

    @classmethod
    def from_roots(cls, tr_ld: TreeLoader, root_offs) -> BranchSet:
        branches = [Branch(tr_ld, root_off) for root_off in root_offs]
        return cls(tr_ld, branches)

    @classmethod
    def from_snap(cls, tr_ld: TreeLoader, root_snap: int,
                  is_c=False,
                  m_s_lb: float = -1.0,
                  m_h_lb: float = -1.0,
                  ) -> BranchSet:
        sel = tr_ld['snap'] == root_snap
        if m_s_lb > 0:
            sel &= tr_ld['m_star'] >= m_s_lb
        if m_h_lb > 0:
            sel &= tr_ld['m_crit200'] >= m_h_lb
        if is_c is not None:
            is_cs = tr_ld['f_in_grp'] == tr_ld['subhalo_id']
            sel &= is_cs == is_c
        root_offs = sel.nonzero()[0]
        return cls.from_roots(tr_ld, root_offs)

    @property
    def size(self):
        return len(self.branches)

    def __len__(self):
        return self.size

    def __getitem__(self, ind):
        return self.branches[ind]

    def root_vals(self, *keys):
        assert len(keys) > 0
        if len(keys) == 1:
            key = keys[0]
            out = np.array([branch.root_val(key) for branch in self.branches])
        else:
            out = tuple(self.root_vals(key) for key in keys)
        return out

    def subset(self, inds):
        return BranchSet(self.tr_ld, [self.branches[ind] for ind in inds])



from __future__ import annotations
from pyhipp.core.abc import HasName, HasSimpleRepr, IsImmutable, HasCache
from pyhipp.field.cubic_box.box import PeriodicBox
from pyhipp.astro.cosmology.model import predefined as predef_cosms, LambdaCDM
from typing import Dict, Any, Union, Literal
from collections.abc import Iterable
from pathlib import Path
from functools import cached_property
import numpy as np, pandas as pd, json, os
from scipy.interpolate import interp1d
from dataclasses import dataclass
import importlib_resources

_PredefinedNames = Literal[
    'tng', 'eagle', 'elucid',
    'elucid_w9l3ha', 'elucid_ext_v2', 'elucid_w9l3ha_ext_v2']

class _Predefined(HasSimpleRepr, HasCache):

    def __init__(self) -> None:
        super().__init__()

        key = 'PYHIPP_SIMS_ROOT_DIR'
        if key in os.environ:
            data_dir = Path(os.environ[key]) / 'data' / 'simulations'
        else:
            _data_dir = importlib_resources.files(
                'pyhipp_sims').joinpath('data/simulations')
            with importlib_resources.as_file(_data_dir) as p:
                data_dir = p
        self.data_dir = data_dir

    def __getitem__(self, name: _PredefinedNames | str) -> SimInfo:
        return self.get_cache_or(name, lambda: self.load(name))

    def to_simple_repr(self):
        c: dict[str, SimInfo] = self.cache
        return {
            'data_dir': str(self.data_dir),
            'models': {
                n: info.to_simple_repr() for n, info in c.items()
            },
        }

    def load(self, name) -> SimInfo:
        return SimInfo.from_conf_file(self.data_dir / f'{name}.json')

    def set_data_dir(self, path: Path) -> None:
        self.data_dir = Path(path)

predefined = _Predefined()

class SimInfo(HasName, HasSimpleRepr, IsImmutable):
    
    predefined = predefined

    @dataclass
    class PartTypeIndex:
        gas:        int = 0
        dm:         int = 1
        tracers:    int = 3
        stars:      int = 4
        bh:         int = 5

    part_type_index = PartTypeIndex()

    def __init__(self,
                 sim_def: Dict[str, Any],
                 root_dir: Path = None,
                 root_file: Path = None,
                 formal_name: str = None,
                 suite: str = None,
                 mahgic_ext: Dict[str, Any] = None,
                 name: str = None) -> None:

        super().__init__(name=name)

        self.sim_def = sim_def
        self.root_dir = Path(root_dir)
        self.root_file = Path(root_file)
        self.formal_name = formal_name
        self.suite = suite
        self.mahgic_ext = mahgic_ext

        self.is_baryon: bool = sim_def['is_baryon']
        self.is_fullbox: bool = sim_def['is_fullbox']
        self.is_fullcover: bool = sim_def['is_fullcover']
        self.partition: Union[None, Dict] = sim_def['partition']
        self.n_snapshots: int = sim_def['n_snapshots']
        self.mass_table = np.array(sim_def['mass_table'])
        self.scale_factors = np.array(sim_def['scale_factors'])
        self.redshifts = np.array(sim_def['redshifts'])
        self.lookback_times = np.array(sim_def['lookback_times'])
        self.box_size: float = float(sim_def['box_size'])
        self.sim_def = sim_def.copy()

        for key in 'scale_factors', 'redshifts', 'lookback_times':
            assert len(getattr(self, key)) == self.n_snapshots

        self.__make_interp()

    @staticmethod
    def from_conf_file(path: Path) -> SimInfo:
        file_name = str(path)
        assert file_name[-5:] == '.json', f'{file_name} must be a JSON file'
        with open(file_name, 'rb') as f:
            conf: dict = json.load(f)

        name: str = conf.pop('name', None)
        formal_name = conf.pop('formal_name', None)
        suite = conf.pop('suite', None)
        root_dir = path.parent / conf.pop('root_dir', None)
        root_file = conf.pop('root_file', None)
        sim_def = conf.pop('simulation_definition')
        mahgic_ext = conf.pop('mahgic_extension', None)

        return SimInfo(sim_def, root_dir=root_dir, root_file=root_file,
                       formal_name=formal_name, suite=suite,
                       mahgic_ext=mahgic_ext, name=name)

    def clone(self) -> SimInfo:
        new_sim_def = self.sim_def.copy()
        p = self.partition
        if p is not None:
            new_sim_def['partition'] = p.copy()
        new_ext = self.mahgic_ext.copy()

        return SimInfo(new_sim_def, root_dir=self.root_dir,
                       root_file=self.root_file, formal_name=self.formal_name,
                       suite=self.suite, mahgic_ext=new_ext,
                       name=self.name)

    def get_sub(self, i_sub: int) -> SimInfo:
        sim_def, p = self.sim_def, self.partition
        assert p is not None
        assert p['subbox_division_policy'] == 'no_overlap'

        sub_sim_def = sim_def.copy()
        sub_sim_def['is_fullbox'] = False
        sub_sim_def['is_fullcover'] = False
        sub_sim_def['partition'] = None
        full_box_size = sim_def['box_size']
        sub_sim_def['box_size'] = full_box_size / p['n_subbox'] ** (1./3.)
        sub_sim_def['full_box_size'] = full_box_size

        name = self.name
        formal_name = self.formal_name
        suite = self.suite
        root_dir = (self.root_dir /
                    p['subboxes_dir'] /
                    (p['subbox_dir_prefix'] + str(i_sub)))
        root_file = p['subbox_root_file']
        mahgix_ext = self.mahgic_ext.copy()

        return SimInfo(
            sub_sim_def, root_dir=root_dir, root_file=root_file,
            formal_name=formal_name, suite=suite, mahgic_ext=mahgix_ext,
            name=name)

    def get_mahgic_extended(self, has_branch_ext=True,
                            has_position_ext=True) -> SimInfo:
        info = self.clone()

        if has_position_ext:
            has_branch_ext = True

        info.mahgic_ext |= {
            'has_branch_ext': has_branch_ext,
            'has_position_ext': has_position_ext,
        }

        return info

    def get_variant(self, sub_dir: str = None) -> SimInfo:
        info = self.clone()
        if sub_dir is not None:
            info.root_dir = self.root_dir / sub_dir

        return info

    @property
    def full_box_size(self):
        if self.is_fullbox:
            return self.box_size
        return self.sim_def['full_box_size']
    
    @property
    def box(self) -> PeriodicBox:
        assert self.is_fullbox
        return PeriodicBox(self.box_size)

    def to_simple_repr(self) -> dict:
        return {
            'name': self.name,
            'formal_name': self.formal_name,
            'root_dir': str(self.root_dir),
            'root_file': str(self.root_file),
            'sim_def': self.sim_def,
            'mahgic_ext': self.mahgic_ext,
        }

    @property
    def extended_root_file(self) -> Path:
        fname = str(self.root_file)

        ext = self.mahgic_ext
        if ext['has_branch_ext']:
            fname += '.bext'
            if ext['has_position_ext']:
                fname += '.pext'

        return self.root_dir / fname

    @cached_property
    def cosmology(self) -> LambdaCDM:
        return predef_cosms[self.sim_def['cosmology']]
    
    @property
    def unit_system(self):
        return self.cosmology.unit_system

    @cached_property
    def ages(self) -> np.ndarray:
        '''
        Cosmic ages [Gyr/h].
        '''
        return self.cosmology.age(self.redshifts)

    @cached_property
    def big_hubbles(self) -> np.ndarray:
        '''
        Hubble parameters H(z) [h/Gyr].
        '''
        return self.cosmology.big_hubble(self.redshifts)

    def z_to_snap(self, z: np.ndarray) -> np.ndarray:
        if isinstance(z, Iterable):
            return np.array([self.z_to_snap(_z) for _z in z], dtype=int)

        idx = np.argmin(np.abs(self.redshifts-z))
        return idx

    def lookback_time_to_z(self, lbt: np.ndarray) -> np.ndarray:
        return self._z_at_lbt(lbt)

    def z_to_lookback_time(self, z: np.ndarray) -> np.ndarray:
        return self._lbt_at_z(z)

    def __make_interp(self):
        zs, lbts = self.redshifts, self.lookback_times
        df = pd.DataFrame({'zs': zs, 'lbts': lbts}).drop_duplicates('zs')
        zs, lbts = df['zs'].to_numpy(), df['lbts'].to_numpy()
        #assert (zs >= 0).all()
        #assert (lbts >= 0).all()
        if zs[-1] > 0.:
            zs, lbts = np.concatenate((zs, (0.,))), \
                np.concatenate((lbts, (0.,)))

        kw = {'kind': 'slinear'}
        self._lbt_at_z = interp1d(zs, lbts, **kw)
        self._z_at_lbt = interp1d(lbts, zs, **kw)



from __future__ import annotations
import numpy as np
from pyhipp.core import abc
from pyhipp.io import h5
from typing import Dict, Callable
from .sim_info import SimInfo
from pyhipp.core.dataproc import Array
from pyhipp.io.binary.fortran import BinaryFile
from functools import cached_property
from dataclasses import dataclass


class TreeLoader(abc.HasLog, abc.HasCache):
    '''
    Provide data fields for the whole tree or a contiguous slice.
    
    Derived class has to define 
    _preproc()
    dgrp_tree
    dataset_key_map
    or optionally, _load()
    
    Available keys
    --------------
    subhalo_id,   f_pro,  n_pro,
    des,main_leaf,
    last_pro, f_in_grp,
    n_in_grp, snap,
    x, v,                               -- cMpc/h, physical km/s, the same for other velocities
    m_star, m_gas, m_bh, m_dm,          -- 10^10 M_sun/h, the same for other masses
    sfr,                                -- M_sun/yr
    m_tophat, r_tophat, m_mean200, ...  -- cMpc/h
    m_v_max, r_vmax,
    m_sub, v_max, v_disp, 
    spin,                               -- (Mpc/h)(km/s)
    ...
                            
    Attrs
    -----
    sim_info, root_file
    '''

    def __init__(self, sim_info: SimInfo):

        super().__init__()

        self.sim_info = sim_info

        self.root_file: h5.File = None
        self.dgrp_tree: h5.Group = None
        self.dataset_key_map: Dict[str, (str | Callable, int, float)] = {}

        self._preproc()
        self._add_mixin()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.root_file.close()

    @staticmethod
    def create(sim_info: SimInfo, **kw) -> TreeLoader:

        cls = {
            'tng': TreeLoaderTng,
            'tng_dark': TreeLoaderTngDark,
            'eagle': TreeLoaderEagle,
            'elucid_ext': TreeLoaderElucidExt,
            'elucid_ext_v2': TreeLoaderElucidExtV2,
            'simba': TreeLoaderSimba,
        }[sim_info.name]

        return cls(sim_info, **kw)

    def __getitem__(self, key) -> np.ndarray | tuple[np.ndarray, ...]:
        if not isinstance(key, str):
            return tuple(self[k] for k in key)
        return self.get_cache_or(key, lambda: self._load(key))
    
    def load_row(self, row: slice|int, *key: str):
        assert len(key) > 0, 'At least one key is required to load a row.'
        if len(key) > 1:
            return tuple(self.load_row(row, _key) for _key in key)
        key, col_slice, scale =  self.dataset_key_map[key[0]]
        return self._load_row(self.dgrp_tree, row, key, col_slice, scale)

    def _preproc(self):
        raise NotImplementedError()

    def _add_mixin(self):
        map = self.dataset_key_map

        def get_is_c():
            return self['f_in_grp'] == self['subhalo_id']
        if 'is_c' not in map:
            map['is_c'] = (get_is_c, None, None)

    def _load(self, key) -> np.ndarray:
        key_used, col_slice, scale = self.dataset_key_map[key]
        return self._load_impl(self.dgrp_tree, key_used,
                               col_slice=col_slice, scale=scale)

    def _load_impl(self, g: h5.Group, key, col_slice=None, scale=None):
        '''
        @key: str | callable | tuple of keys. 
        For the last, col_slice and scale are applied to the value of each 
        key; returned values are stacked.
        '''
        if isinstance(key, str):
            d = g[key]
            val = d[()] if col_slice is None else d[:, col_slice]
            if scale is not None:
                val *= scale
        elif isinstance(key, Callable):
            val = key()
            if col_slice is not None:
                val = val[:, col_slice]
            if scale is not None:
                val *= scale
        else:
            val = np.stack(
                tuple(self._load_impl(g, k, col_slice, scale) for k in key),
                axis=-1)
        return val
    
    def _load_row(self, g: h5.Group, row: slice|int, 
                  key: str, col_slice=None, scale=None):
        if isinstance(key, str):
            dset = g[key]
            val = dset[row] if col_slice is None else dset[row, col_slice]    
            if scale is not None:
                val *= scale
            return val

class TreeLoaderTngDark(TreeLoader):

    def _preproc(self):
        info = self.sim_info
        self.root_file = h5.File(info.root_dir / info.root_file)
        self.dgrp_tree = self.root_file['Trees/SubLink']
        self.dataset_key_map = {
            'subhalo_id':           ('SubhaloID', None, None),
            'subfind_id':           ('SubfindID', None, None),

            'f_pro':                ('FirstProgenitorID', None, None),
            'n_pro':                ('NextProgenitorID', None, None),
            'des':                  ('DescendantID', None, None),
            'main_leaf':            ('MainLeafProgenitorID', None, None),
            'last_pro':             ('LastProgenitorID', None, None),
            'f_in_grp':             ('FirstSubhaloInFOFGroupID', None, None),
            'n_in_grp':             ('NextSubhaloInFOFGroupID', None, None),
            'snap':                 ('SnapNum', None, None),
            'x':                    ('SubhaloPos', None, 1.0e-3),
            'v':                    ('SubhaloVel', None, None),
            'm':                    ('SubhaloMassInRad', None, None),
            'm_sub':                ('SubhaloMass', None, None),
            'm_dm_sub':             ('SubhaloMassType', 1, None),
            'm_dm':                 ('SubhaloMassInRadType', 1, None),
            'm_dm_v_max':           ('SubhaloMassInMaxRadType', 1, None),

            'v_max':                ('SubhaloVmax', None, None),
            'r_v_max':              ('SubhaloVmaxRad', None, 1.0e-3),
            'm_v_max':              ('SubhaloMassInMaxRad', None, None),
            'v_disp':               ('SubhaloVelDisp', None, None),
            'spin':                 ('SubhaloSpin', None, 1.0e-3),

            'm_tophat':             ('Group_M_TopHat200', None, None),
            'm_crit200':            ('Group_M_Crit200', None, None),
            'm_mean200':            ('Group_M_Mean200', None, None),
            
            'r_tophat':             ('Group_R_TopHat200', None, 1.0e-3),
            'r_crit200':            ('Group_R_Crit200', None, 1.0e-3),
            'r_mean200':            ('Group_R_Mean200', None, 1.0e-3),
        }


class TreeLoaderTng(TreeLoaderTngDark):

    def _preproc(self):
        super()._preproc()

        self.dataset_key_map |= {
            'sfr':                  ('SubhaloSFRinRad', None, None),
            'sfr_sub':              ('SubhaloSFR', None, None),

            'r_half_mass_star':     ('SubhaloHalfmassRadType', 4, 1.0e-3),
            'm_star':               ('SubhaloMassInRadType', 4, None),
            'm_gas':                ('SubhaloMassInRadType', 0, None),
            'm_bh':                 ('SubhaloMassInRadType', 5, None),
            
            'n_ps_star':             ('SubhaloLenType', 4, None),
            'n_ps_gas':             ('SubhaloLenType', 0, None),
            'n_ps_bh':             ('SubhaloLenType', 5, None),

            'm_star_sub':           ('SubhaloMassType', 4, None),
            'm_gas_sub':            ('SubhaloMassType', 0, None),
            'm_bh_sub':             ('SubhaloMassType', 5, None),

            'm_star_v_max':         ('SubhaloMassInMaxRadType', 4, None),
            'm_gas_v_max':          ('SubhaloMassInMaxRadType', 0, None),
            'm_bh_v_max':           ('SubhaloMassInMaxRadType', 5, None),
        }


class TreeLoaderEagle(TreeLoader):
    def _preproc(self):
        info = self.sim_info
        self.root_file = h5.File(info.root_dir / info.root_file)
        self.dgrp_tree = self.root_file['Trees/DTrees']

        h = info.cosmology.hubble
        m_scale = 1.0e-10 * h
        x_scale = h

        self.dataset_key_map = {
            'subhalo_id': ('GalaxyID', None, None),
            'f_pro': ('FirstProgenitorID', None, None),
            'n_pro': ('NextProgenitorID', None, None),
            'des': ('DescendantID', None, None),
            'main_leaf': ('TopLeafID', None, None),
            'last_pro': ('LastProgID', None, None),
            'f_in_grp': ('FirstSubhaloInFOFGroupID', None, None),
            'snap': ('SnapNum', None, None),
            'x': (('CentreOfPotential_x', 'CentreOfPotential_y', 'CentreOfPotential_z'), None, x_scale),
            'v': (('Velocity_x', 'Velocity_y', 'Velocity_z')),
            'm_star': ('Mass_Star_Aperture30', None, m_scale),
            'm_gas': ('Mass_Gas_Aperture30', None, m_scale),
            'm_bh': ('Mass_BH_Aperture30', None, m_scale),
            'm_dm': ('Mass_DM_Aperture30', None, m_scale),
            'sfr': ('SFR_Aperture30', None, None),
            
            'm_tophat': ('Group_M_TopHat200', None, m_scale),
            'm_crit200': ('Group_M_Crit200', None, m_scale),
            'm_mean200': ('Group_M_Mean200', None, m_scale),
            
            'm_sub': ('Mass', None, m_scale),
            'v_max': ('Vmax', None, None),
            'v_disp': ('VelDisp_Aperture30', None, None),
        }


class TreeLoaderSimba(TreeLoader):
    def _preproc(self):
        info = self.sim_info

        self.root_file = h5.File(info.root_dir / info.root_file)
        self.dgrp_tree = self.root_file['Trees/descend_galaxy_star/Galaxies']

        h = info.cosmology.hubble
        m_scale = 1.0e-10 * h
        x_scale = 1.0e-3 * h

        self.dataset_key_map = {
            'subhalo_id': ('id', None, None),
            'f_pro': ('f_pro', None, None),
            'n_pro': ('n_pro', None, None),
            'des': ('des', None, None),
            'main_leaf': ('main_leaf', None, None),
            'last_pro': ('last_pro', None, None),
            'f_in_grp': ('f_in_grp', None, None),
            'n_in_grp': ('n_in_grp', None, None),
            'snap': ('snap', None, None),
            'x': ('minpotpos', None, x_scale),
            'v': ('minpotvel', None, None),
            'm_star': ('mass_stellar', None, m_scale),
            'm_gas': ('mass_gas', None, m_scale),
            'm_bh': ('mass_bh', None, m_scale),
            'sfr': ('sfr', None, None),
            'm_tophat': ('halo_m200c', None, m_scale),
        }


class TreeLoaderElucidRaw:

    subhalo_dtype = np.dtype([
        ('des', np.int32),
        ('f_pro', np.int32),
        ('n_pro', np.int32),
        ('f_in_grp', np.int32),
        ('n_in_grp', np.int32),
        ('len', np.int32),
        ('m_mean200', np.float32),
        ('m_crit200', np.float32),
        ('m_tophat', np.float32),
        ('x', np.float32, 3),
        ('v', np.float32, 3),
        ('v_disp', np.float32),
        ('v_max', np.float32),
        ('spin', np.float32, 3),
        ('most_bound_pid', np.int64),
        ('snap', np.int32),
        ('file_id', np.int32),
        ('subfind_id', np.int32),
        ('r_sub_half_mass', np.float32),
    ], align=True)

    @dataclass
    class TreeData:
        n_trees: int
        n_subhs: int
        n_subhs_of_trees: np.ndarray
        subhs: np.ndarray
        meta: dict

    def __init__(self, chunk_id=0, sim_info=SimInfo.predefined['elucid'],
                 data_dir='raw/postproc/treedata'):
        self.chunk_id = chunk_id
        self.sim_info = sim_info
        self.data_dir = data_dir

    def __getitem__(self, key):
        if not isinstance(key, str):
            return tuple(self[k] for k in key)
        td = self._tree_data
        if key in td.meta:
            val = td.meta[key]
        else:
            val = td.subhs[key]
        return val

    @cached_property
    def _tree_data(self):
        info = self.sim_info
        snap = info.n_snapshots - 1
        c = self.chunk_id
        file_path = info.root_dir / self.data_dir / f'trees_{snap}.{c}'
        with BinaryFile(file_path) as f:
            n_trees = int(f.load_rec(np.int32))
            n_subhs = int(f.load_rec(np.int32))
            n_subhs_of_trees = f.load_rec(np.int32, n_trees)
            subhs = f.load_rec(self.subhalo_dtype, n_subhs)
            assert len(f.load_bytes(1)) == 0
        
        offset_of_trees = Array.count_to_offset(n_subhs_of_trees)
        assert n_subhs == offset_of_trees[-1]
        subhalo_id = np.zeros(n_subhs, dtype=np.int32)
        tree_id = np.zeros(n_subhs, dtype=np.int32)
        for i, (b, e) in enumerate(
            zip(offset_of_trees[: -1],
                offset_of_trees[1:])):
            subhalo_id[b:e] = np.arange(e-b)
            tree_id[b:e] = i
        is_c = subhalo_id == subhs['f_in_grp']
        meta = {
            'offset_of_trees': offset_of_trees,
            'subhalo_id': subhalo_id,
            'tree_id': tree_id,
            'is_c': is_c,
        }
            
        return self.TreeData(n_trees, n_subhs, n_subhs_of_trees, subhs, meta)


class TreeLoaderElucid(TreeLoader):
    def __init__(self, file_id: int, *, sim_info=SimInfo.predefined['elucid']):

        self.file_id = file_id

        super().__init__(sim_info)

    def _preproc(self):
        info, file_id = self.sim_info, self.file_id

        root_file = h5.File(info.root_dir / info.root_file)
        dgrp = root_file[f'Trees/SubLink/{file_id}']
        dgrp_tree = dgrp[f'Subhalos']
        n_halos_in_trees = dgrp['Header'].datasets['NumHalosInTree']
        offset_of_trees = Array.count_to_offset(n_halos_in_trees)

        n_h_tot = n_halos_in_trees.sum()
        assert n_h_tot == offset_of_trees[-1]
        subhalo_id = np.zeros(n_h_tot, dtype=np.int32)
        tree_id = np.zeros(n_h_tot, dtype=np.int32)
        for i, (b, e) in enumerate(
            zip(offset_of_trees[: -1],
                offset_of_trees[1:])):
            subhalo_id[b:e] = np.arange(e-b)
            tree_id[b:e] = i

        def get_subhalo_id():
            return subhalo_id

        def get_tree_id():
            return tree_id

        dataset_key_map = {
            'tree_id': (get_tree_id, None, None),
            'subhalo_id': (get_subhalo_id, None, None),         # = id_in_tree
            'id_in_tree': (get_subhalo_id, None, None),

            'des': ('Descendant', None, None),
            'f_pro': ('FirstProgenitor', None, None),
            'n_pro': ('NextProgenitor', None, None),
            'f_in_grp': ('FirstHaloInFOFgroup', None, None),
            'n_in_grp': ('NextHaloInFOFgroup', None, None),
            'len': ('Len', None, None),
            'm_tophat': ('M_TopHat', None, None),
            'm_crit200': ('M_Crit200', None, None),
            'm_mean200': ('M_Mean200', None, None),
            'x': ('Pos', None, None),
            'v': ('Vel', None, None),
            'v_disp': ('VelDisp', None, None),
            'v_max': ('Vmax', None, None),
            'spin': ('Spin', None, None),
            'most_bound_pid': ('MostBoundID', None, None),
            'snap': ('SnapNum', None, None),
            'file_id': ('FileNr', None, None),
            'subfind_id': ('SubhaloIndex', None, None),
            'r_sub_half_mass': ('SubhalfMass', None, None),     # pMpc/h
        }

        self.root_file = root_file
        self.dgrp_tree = dgrp_tree
        self.n_halos_in_trees = n_halos_in_trees
        self.offset_of_trees = offset_of_trees
        self.dataset_key_map = dataset_key_map


class TreeLoaderElucidExt(TreeLoader):
    def _preproc(self):

        info = self.sim_info
        self.root_file = h5.File(info.extended_root_file)
        self.dgrp_tree = self.root_file['Trees/SubLink/Subhalos']

        self.dataset_key_map = {
            'subhalo_id': ('subhalo_id', None, None),
            'id_in_tree': ('id_in_tree', None, None),
            'tree_id': ('tree_id', None, None),

            'chunk_id': ('chunk_id', None, None),
            'src_chunk_id': ('src_chunk_id', None, None),
            'src_id_in_tree': ('src_id_in_tree', None, None),
            'src_tree_id': ('src_tree_id', None, None),

            'snap': ('snap', None, None),
            'f_pro': ('f_pro', None, None),
            'n_pro': ('n_pro', None, None),
            'des': ('des', None, None),
            'main_leaf': ('main_leaf', None, None),
            'f_in_grp': ('f_in_grp', None, None),
            'last_pro': ('last_pro', None, None),
            'n_in_grp': ('n_in_grp', None, None),

            'x': ('x', None, None),
            'v': ('v', None, None),

            'm_tophat': ('m_tophat', None, None),
            'm_crit200': ('m_crit_200', None, None),
            'm_mean200': ('m_mean_200', None, None),
            'sub_half_mass': ('sub_half_mass', None, None),
            'r_tophat': ('r_tophat', None, None),
            'v_max': ('v_max', None, None),

            'most_bound_pid': ('most_bound_pid', None, None),
            'spin': ('spin', None, None),
            'vel_disp': ('vel_disp', None, None),
        }


class TreeLoaderElucidExtV2(TreeLoader):
    def _preproc(self):

        info = self.sim_info
        has_ext = info.mahgic_ext['has_position_ext']

        file_name = 'sublink.hdf5'
        if has_ext:
            file_name += '.ext'
        root_file = h5.File(info.root_dir/file_name)
        dgrp_tree = root_file['Subhalos']

        def get_m_sub():
            ind_dm = info.part_type_index.dm
            m_sub = self['len'] * info.mass_table[ind_dm]
            return m_sub

        dataset_key_map = {
            'des': ('des', None, None),
            'f_in_grp': ('f_in_grp', None, None),
            'f_pro': ('f_pro', None, None),
            'last_pro': ('last_pro', None, None),
            'len': ('len', None, None),
            'm_crit_200': ('m_crit_200', None, None),
            'm_mean_200': ('m_mean_200', None, None),

            'm_crit200': ('m_crit_200', None, None),    # alias
            'm_mean200': ('m_mean_200', None, None),    # alias

            'm_tophat': ('m_tophat', None, None),
            'main_leaf': ('main_leaf', None, None),
            'most_bound_pid': ('most_bound_pid', None, None),
            'n_in_grp': ('n_in_grp', None, None),
            'n_pro': ('n_pro', None, None),
            'r_tophat': ('r_tophat', None, None),
            'snap': ('snap', None, None),
            'spin': ('spin', None, None),
            'sub_half_mass': ('sub_half_mass', None, None),
            'subhalo_id': ('subhalo_id', None, None),
            'v': ('v', None, None),
            'v_max': ('v_max', None, None),
            'vel_disp': ('vel_disp', None, None),
            'x': ('x', None, None),
            'm_sub': (get_m_sub, None, None),
        }

        if has_ext:
            dataset_key_map |= {
                'm_sub': ('m_sub', None, None),
                'x_ext': ('x_ext', None, None),
                'v_ext': ('v_ext', None, None),
                'is_kept': ('is_kept', None, None),
                'src_flag': ('src_flag', None, None),
                'src_subhalo_id': ('src_subhalo_id', None, None),
            }
        else:
            dataset_key_map |= {
                'chunk_id': ('chunk_id', None, None),
                'tree_id': ('tree_id', None, None),
                'id_in_tree': ('id_in_tree', None, None),
                'src_chunk_id': ('src_chunk_id', None, None),
                'src_tree_id': ('src_tree_id', None, None),
                'src_id_in_tree': ('src_id_in_tree', None, None),
            }

        self.root_file = root_file
        self.dgrp_tree = dgrp_tree
        self.dataset_key_map = dataset_key_map



from __future__ import annotations
from typing import Any
from .sim_info import SimInfo
from pyhipp.io import h5
from pyhipp.io.binary.fortran import BinaryFile
from pyhipp.core import abc
from typing import Tuple
import numpy as np
from functools import cached_property


class SnapshotLoader(abc.HasLog):

    part_types: Tuple[str, ...] = ()

    def __init__(self, sim_info: SimInfo, z: float = None,
                 snap: int = None, **kw) -> None:
        super().__init__(**kw)

        if z is None:
            assert snap is not None
        else:
            assert snap is None
            snap = sim_info.z_to_snap(z)

        self.sim_info = sim_info
        self.snap = snap
        self._root_file: h5.File = None

    @property
    def root_file(self) -> h5.File:
        if self._root_file is None:
            self._root_file = self._open_root_file()
        return self._root_file
    
    def _open_root_file(self):
        info = self.sim_info
        path = info.root_dir / info.root_file
        return h5.File(path)

    def iter_particles(self, part_type='dm', keys=('x', )):
        raise NotImplementedError()
    
    def load_row(self, part_type: str, row: slice|int, *key: str):
        assert len(key) > 0
        if len(key) > 1:
            return tuple(self.load_row(part_type, row, _key) for _key in key)
        return self._load_row(part_type, row, key[0])
    
    def _load_row(self, part_type: str, row: slice|int, key: str) -> np.ndarray | Any:
        raise NotImplementedError()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        if self._root_file is not None:
            self._root_file.close()
            self._root_file = None
        

class SnapshotLoaderDmo(SnapshotLoader):

    part_types = ('dm', )

    @property
    def part_mass_dm(self):
        raise NotImplementedError()

    @property
    def n_part_dm(self):
        raise NotImplementedError()

    @staticmethod
    def create(sim_type, **kw):
        return {
            'tng_dark': SnapshotLoaderTngDark,
        }[sim_type](**kw)


class SnapshotLoaderTngDark(SnapshotLoaderDmo):

    part_type_to_id = {
        'dm': 1,
    }

    # dataset name and scale to standard units
    prop_key_to_meta = {
        'x': ('Coordinates', 1.0e-3),
        'v': ('Velocities', None),
        'potential': ('Potential', None),
        'id': ('ParticleIDs', None),
    }

    def __init__(self, sim_info: SimInfo, z: float = None,
                 snap: int = None, **kw) -> None:
        super().__init__(sim_info, z, snap, **kw)

    @property
    def part_mass_dm(self):
        part_type = self.part_type_to_id['dm']
        return self.sim_info.mass_table[part_type]

    @property
    def n_part_dm(self):
        part_type = self.part_type_to_id['dm']
        f = self.root_file
        dgrp_snap = f[f'Snapshots/{self.snap}']
        n_ps = dgrp_snap['Header'].attrs['NumPart_Total'][part_type]
        return n_ps

    def iter_particles(self, part_type='dm', keys=('x', ),
                       *, n_chunk=1024*1024*4, n_min=0, n_max=None):
        '''
        @chunk_size: None for load all at once, int for the number of particles
        loaded at every iteration.
        '''
        part_type = self.part_type_to_id[part_type]
        metas = tuple(self.prop_key_to_meta[k] for k in keys)

        f = self.root_file
        dgrp_snap = f[f'Snapshots/{self.snap}']
        dgrp_ps = dgrp_snap[f'PartType{part_type}']
        n_ps = dgrp_snap['Header'].attrs['NumPart_Total'][part_type]
        dsets = tuple(dgrp_ps[k] for k, _ in metas)

        if n_max is None:
            n_max = n_ps
        if n_chunk is None:
            n_chunk = n_max - n_min

        for b in range(n_min, n_max, n_chunk):
            e = min(b + n_chunk, n_max)
            yield tuple(dset[b:e] if scale is None else dset[b:e]*scale
                        for dset, (_, scale) in zip(dsets, metas))


class SnapshotLoaderElucidRaw(SnapshotLoaderDmo):
    '''
    Load data in raw binary format.
    Instance is immutable.
    '''

    header_dtype = np.dtype([
        ('NumPart_ThisFile', np.uint32, 6),
        ('MassTable', np.float64, 6),
        ('Time', np.float64),
        ('Redshift', np.float64),
        ('FlagSFR', np.int32),
        ('FlagFeedback', np.int32),
        ('NumPart_Total', np.uint32, 6),
        ('FlagCooling', np.int32),
        ('NumFilesPerSnapshot', np.int32),
        ('BoxSize', np.float64),
        ('Omega0', np.float64),
        ('OmegaLambda', np.float64),
        ('HubbleParam', np.float64),
        ('FlagStellarAge', np.int32),
        ('FlagMetals', np.int32),
        ('HashTableSize', np.int32),
        ('fill', 'S84'),
    ], align=True)
    fields = [
        ('x', np.float32, (3,)),
        ('v', np.float32, (3,)),
        ('id', np.int64, ()),
    ]
    n_files = 2048
    p_type_dm = 1

    def __init__(self, sim_info: SimInfo, z: float = None, snap: int = None,
                 chunk_id=0, **kw) -> None:
        super().__init__(sim_info, z, snap, **kw)

        assert 0 <= chunk_id < self.n_files
        assert self.header_dtype.itemsize == 256
        self.chunk_id = chunk_id

    @property
    def root_file_name(self):
        snap, c = self.snap, self.chunk_id
        return self.sim_info.root_dir / f'raw/snapdir_{snap:03d}/snapshot_{snap:03d}.{c}'

    def _open_root_file(self):
        return BinaryFile(self.root_file_name, 'rb')

    @cached_property
    def header(self):
        f = self.root_file
        header = f.load_sect(self.header_dtype)
        return header

    @property
    def part_mass_dm(self):
        return float(self.header['MassTable'][self.p_type_dm])

    @property
    def n_part_dm(self):
        return int(self.header['NumPart_ThisFile'][self.p_type_dm])

    def load_particles(self, field='x'):
        n_p = self.n_part_dm
        f = self.root_file
        f.load_sect(self.header_dtype)
        val = None
        for _field, _dtype, _shape in self.fields:
            _dt = np.dtype(_dtype)
            full_shape = (n_p,) + _shape
            n = np.prod(full_shape)
            if _field != field:
                f.skip_sect(_dt, n=n)
            else:
                val = f.load_sect(_dt, n=n).reshape(full_shape)
        assert val is not None, f'Field {field} not found.'
        return val
    
    @cached_property
    def hash_table(self):
        n_p = self.n_part_dm
        f = self.root_file
        f.skip_sect(self.header_dtype)
        for field, dtype, shape in self.fields:
            _dtype = np.dtype(dtype)
            full_shape = (n_p,) + shape
            n = np.prod(full_shape)
            f.skip_sect(_dtype, n=n)
        bcell, ecell = f.load_sect(np.int32, 2)
        n_cells = ecell - bcell + 1
        p_offs = f.load_sect(np.int32, n_cells)
        return {
            'FirstCellID': bcell,
            'LastCellID': ecell,
            'ParticleOffsetsInCell': p_offs,
        }

class _SnapshotLoaderElucidChunk(SnapshotLoaderDmo):
    part_type_to_id = {
        'dm': 1,
    }
    prop_key_to_meta = {
        'x': ('Coordinates', None),
        'v': ('Velocities', None),
        'id': ('ParticleIDs', None),
    }

    def __init__(self, sim_info: SimInfo, z: float = None, snap: int = None,
                 chunk_id=0, **kw) -> None:
        super().__init__(sim_info, z, snap, **kw)

        self.chunk_id = chunk_id

    @property
    def part_mass_dm(self):
        part_type = self.part_type_to_id['dm']
        return self.sim_info.mass_table[part_type]

    @property
    def n_part_dm(self):
        part_type = self.part_type_to_id['dm']
        f = self.root_file
        dgrp_ps = f[f'Snapshots/{self.snap}/{self.chunk_id}/PartType{part_type}']
        n_ps = dgrp_ps.attrs['NumPart_ThisFile']
        return n_ps

    def iter_particles(self, part_type='dm', keys=('x', ), *,
                       n_chunk=1024 * 1024 * 4, n_min=0, n_max=None):
        part_type = self.part_type_to_id['dm']
        metas = tuple(self.prop_key_to_meta[k] for k in keys)

        f = self.root_file
        dgrp_ps = f[f'Snapshots/{self.snap}/{self.chunk_id}/PartType{part_type}']
        n_ps = dgrp_ps.attrs['NumPart_ThisFile']
        dsets = tuple(dgrp_ps[k] for k, _ in metas)

        if n_max is None:
            n_max = n_ps
        if n_chunk is None:
            n_chunk = n_max - n_min

        for b in range(n_min, n_max, n_chunk):
            e = min(b + n_chunk, n_max)
            yield tuple(dset[b:e] if scale is None else dset[b:e]*scale
                        for dset, (_, scale) in zip(dsets, metas))


class SnapshotLoaderElucid(SnapshotLoaderDmo):
    def __init__(self, sim_info: SimInfo, z: float = None,
                 snap: int = None, **kw) -> None:
        super().__init__(sim_info, z, snap, **kw)

    @property
    def part_mass_dm(self):
        part_type = _SnapshotLoaderElucidChunk.part_type_to_id['dm']
        return self.sim_info.mass_table[part_type]

    @property
    def n_part_dm(self):
        part_type = _SnapshotLoaderElucidChunk.part_type_to_id['dm']
        f = self.root_file
        n_ps = f[f'Header'].attrs['NumPart_Total'][part_type]
        return n_ps

    @property
    def n_chunks(self):
        f = self.root_file
        n_c = f[f'Header'].attrs['NumFilesPerSnapshot']
        return n_c

    def chunk_loader(self, chunk_id: int):
        return _SnapshotLoaderElucidChunk(
            sim_info=self.sim_info, snap=self.snap, chunk_id=chunk_id)

    def iter_particles(self, part_type='dm', keys=('x', ), *,
                       chunks: list[int] = None, chunk_loader_kw={}):
        if chunks is None:
            chunks = np.arange(self.n_chunks)
        for c in chunks:
            yield from self.chunk_loader(c)\
                .iter_particles(part_type, keys, **chunk_loader_kw)


class SnapshotLoaderBaryonic(SnapshotLoader):

    part_types = ('gas', 'dm', 'star', 'bh')

    @property
    def part_mass_gas(self):
        raise NotImplementedError()

    @property
    def part_mass_dm(self):
        raise NotImplementedError()

    @property
    def part_mass_star(self):
        raise NotImplementedError()

    @property
    def part_mass_bh(self):
        raise NotImplementedError()

    @staticmethod
    def create(sim_type, **kw):
        return {
            'tng': SnapshotLoaderTng,
        }[sim_type](**kw)


class SnapshotLoaderTng(SnapshotLoaderBaryonic):

    part_type_to_id = {
        'gas': 0,
        'dm': 1,
        'star': 4,
        'bh': 5,
    }

    prop_key_to_meta = {
        'x': ('Coordinates', 1.0e-3),
        'v': ('Velocities', None),
        'potential': ('Potential', None),
        'id': ('ParticleIDs', None),
        'mass': ('Masses', None),
        'birth_x': ('BirthPos', 1.0e-3),
        'birth_v': ('BirthVel', None),
        'birth_a': ('GFM_StellarFormationTime', None),
        'birth_mass': ('GFM_InitialMass', None)
    }

    @property
    def part_mass_dm(self):
        part_type = self.part_type_to_id['dm']
        return self.sim_info.mass_table[part_type]

    def iter_particles(self, part_type: str, keys=('x', ),
                       n_chunks=1024*1024*1024, n_max=None):
        part_type = self.part_type_to_id[part_type]
        metas = tuple(self.prop_key_to_meta[k] for k in keys)

        f = self.root_file
        dgrp_snap = f[f'Snapshots/{self.snap}']
        dgrp_ps = dgrp_snap[f'PartType{part_type}']
        n_ps = dgrp_snap['Header'].attrs['NumPart_Total'][part_type]
        dsets = tuple(dgrp_ps[k] for k, _ in metas)

        if n_max is None:
            n_max = n_ps
        if n_chunks is None:
            n_chunks = n_max

        for b in range(0, n_max, n_chunks):
            e = min(b + n_chunks, n_max)
            yield tuple(dset[b:e] if scale is None else dset[b:e]*scale
                        for dset, (_, scale) in zip(dsets, metas))
            
    @cached_property
    def dgrp_snap(self):
        return self.root_file[f'Snapshots/{self.snap}']

    @cached_property
    def dgrp_part_types(self):
        return {k: self.dgrp_snap[f'PartType{id}'] 
                for k, id in self.part_type_to_id.items()}
    
    def _load_row(self, part_type: str, row: slice|int, key: str) -> np.ndarray | Any:
        dgrp_ps = self.dgrp_part_types[part_type]
        key, scale = self.prop_key_to_meta[key]
        dset = dgrp_ps[key]
        val = dset[row]
        if scale is not None:
            val *= scale
        return val